from typing import Dict, List

import CloudFlare
import CloudFlare.exceptions

from anemoi.providers import Provider
from anemoi.util import anlog, is_ip_record_valid


class CloudflareProvider(Provider):
    API: CloudFlare.CloudFlare = None

    # parse config
    def __init__(self, config):
        try:  # noqa: SIM105
            if token := config.get("token"):
                self.API = CloudFlare.CloudFlare(token=token)
            elif (email := config.get("email")) and (key := config.get("key")):
                self.API = CloudFlare.CloudFlare(email=email, key=key)
        except CloudFlare.exceptions.CloudFlareAPIError:
            return None

    def __get_zone_for_subdomain(self, subdomain):
        zone = ".".join(subdomain.split(".")[-2:])
        try:
            zones = self.API.zones.get(params={"name": zone})
        except CloudFlare.exceptions.CloudFlareAPIError as e:
            anlog.error(e)
            return None
        if len(zones) != 1:
            anlog.error("Could not find zone")
            return None
        return zones[0]

    def __get_records(self, subdomain) -> tuple[List[str], str]:
        zone = self.__get_zone_for_subdomain(subdomain)
        zid = zone["id"]
        try:
            recs = self.API.zones.dns_records.get(
                zid, params={"name": subdomain, "match": "all"}
            )
        except CloudFlare.exceptions.CloudFlareAPIError as e:
            anlog.error(e)
            return None, zid
        if len(recs) == 0:
            anlog.info(f"No records for {subdomain} found!")
            return None, zid
        # later, maybe implement something to follow CNAMEs within the same domain
        # might want to have an explicit switch for that though
        return [x for x in recs if x["type"] in ("A", "AAAA")], zid

    # returns list of {'A': '1.1.1.1'} objects
    def get_record_ips(self, subdomain) -> List[Dict[str, str]]:
        recs, _ = self.__get_records(subdomain)
        if recs:
            return [{x["type"]: x["content"]} for x in recs]
        return []

    # returns bool of if the update succeeded or not
    def update_record_ip(self, subdomain, ip, rtype="A", proxied=False) -> bool:
        if not is_ip_record_valid(ip, rtype):
            return False

        recs, zid = self.__get_records(subdomain)
        if recs:
            recs = [x for x in recs if x["type"] == rtype]
        if not recs:  # create new record
            try:
                self.API.zones.dns_records.post(
                    zid, data={"name": subdomain, "type": rtype, "content": ip}
                )
            except CloudFlare.exceptions.CloudFlareAPIError as e:
                anlog.error(e)
                return False
            return True

        for rec in recs:
            if ip == rec["content"]:
                continue  # dont update if we dont have to
            try:
                self.API.zones.dns_records.put(
                    zid,
                    rec["id"],
                    data={
                        "name": subdomain,
                        "type": rec["type"],
                        "content": ip,
                        "proxied": rec["proxied"],
                    },
                )
            except CloudFlare.exceptions.CloudFlareAPIError as e:
                anlog.error(e)
                return False
        return True



import importlib
from typing import Dict, List, Optional

from anemoi.client import Client


class Backend:
    def __init__(self, config: Dict):
        # do something with your {'type':'aaa', 'vendor': 'bbb', 'path': 'ccc'} config here
        pass

    def add_client(self, client: Client):
        pass

    # return UUID if success, None if fail
    def delete_client(self, client: Client) -> Optional[str]:
        return None

    # return Client() object if success, None if fail
    def get_client(
        self, uuid: Optional[str] = None, domain: Optional[str] = None
    ) -> Optional[Client]:
        return None

    def update_ip(self, client: Client, ip: str, version: int):
        pass

    @property
    def clients(self) -> List[Client]:
        return []


def init_backend(
    config: Dict,
    default_type="tinydb",
    default_vendor="tinydb",
    default_path="~/.anemoi/clients.json",
) -> Backend:
    default_backend = {
        "type": default_type,
        "vendor": default_vendor,
        "path": default_path,
    }
    backend_config = default_backend | config.get("backend", default_backend)
    if backend_config["type"] == "database" and backend_config["vendor"] == "tinydb":
        raise Exception(
            f"Incompatible backend type and vendor. Make sure vendor is correct for the type '{backend_config['type']}"
        )
    backend_obj: Optional[Backend] = getattr(
        importlib.import_module(f"anemoi.backends.{backend_config['type']}"),
        f"{backend_config['type'].capitalize()}Backend",
    )
    if not backend_obj:
        raise ModuleNotFoundError(
            f"No backend named {backend_config['vendor'].capitalize()}"
        )
    return backend_obj(backend_config)



from typing import Dict, List

import requests

from anemoi.providers import Provider
from anemoi.util import anlog, is_ip_record_valid


class PorkbunProvider(Provider):
    # https://porkbun.com/api/json/v3/documentation
    uri = "https://api.porkbun.com/api/json/"
    version = "v3"
    key: str = None
    secret: str = None

    def __init__(self, config):
        if (apikey := config.get("apikey")) and (secret := config.get("secret")):
            self.key = apikey
            self.secret = secret
        else:
            anlog.error("Insufficient credentials for Porkbun")
            return None

    def _post(self, endpoint, data=None):
        if not data:
            data = {}
        data.update({"apikey": self.key, "secretapikey": self.secret})
        res = requests.post(f"{self.uri}{self.version}/{endpoint}", json=data)
        if res.status_code != 200:
            try:
                if (output := res.json()) and output.get("status", "") == "ERROR":
                    raise Exception(
                        f"Error {res.status_code} on Porkbun API: {output.get('message','unknown')}"
                    )
            except requests.exceptions.JSONDecodeError:
                raise Exception(f"Unknown error {res.status_code} on Porkbun API.")
        data = res.json()
        if data.get("status", "") != "SUCCESS":
            raise Exception("Unknown error in Porkbun API")
        data.pop("status", None)
        return data

    def __get_records(self, subdomain) -> List[str]:
        domain = ".".join(subdomain.split(".")[-2:])
        recs = []
        try:
            res = self._post(f"dns/retrieve/{domain}")
            raw_records = res.get("records", [])
            targets = [x for x in raw_records if x.get("name", "") == subdomain]
            if len(targets) != 0:
                recs = targets
        except Exception as e:
            anlog.error(e)
        return recs

    # returns list of {'A': '1.1.1.1'} objects
    def get_record_ips(self, subdomain) -> List[Dict[str, str]]:
        result = []
        recs = self.__get_records(subdomain)
        for rec in recs:
            if (kind := rec.get("type")) and (ip := rec.get("content")):
                result.append({kind: ip})
        return result

    # returns bool of if the update succeeded or not
    def update_record_ip(self, subdomain: str, ip, rtype="A") -> bool:
        if not is_ip_record_valid(ip, rtype):
            return False
        domain = ".".join(subdomain.split(".")[-2:])
        recs = self.__get_records(subdomain)
        name = "" if subdomain.count(".") <= 1 else subdomain.split(".", 1)[0]
        try:
            if recs:
                recs = [x for x in recs if x.get("type", "") == rtype]
            else:
                # need to create record
                res = self._post(
                    f"dns/create/{domain}", {"name": name, "type": rtype, "content": ip}
                )
                if "id" not in res:
                    anlog.error(f"Failed to create record for {subdomain}")
                    return False
                return True
            for rec in recs:
                if ip == rec.get("content", ""):
                    continue
                res = self._post(
                    f"dns/editByNameType/{domain}/{rtype}/{name}", {"content": ip}
                )
                return True
        except Exception as e:
            anlog.error(e)
        return False



from dataclasses import asdict
from typing import Dict, Optional

from tinydb import Query, TinyDB

from anemoi.client import Client

from . import Backend


class TinydbBackend(Backend):
    db: TinyDB

    def __init__(self, config: Dict):
        self.db = TinyDB(config.get("path"))

    def add_client(self, client: Client):
        self.db.insert(asdict(client))

    def delete_client(self, client: Client):
        client_query = Query()
        res = self.db.search(client_query.domain == client.domain)
        if len(res) == 1:
            client = Client(**res[0])
            res = self.db.remove(doc_ids=[res[0].doc_id])
            if len(res) == 1:
                return client.uuid
        return None

    def get_client(self, uuid=None, domain=None) -> Optional[Client]:
        client_query = Query()
        res = []
        if uuid:
            res = self.db.search(client_query.uuid == uuid)
        elif domain:
            res = self.db.search(client_query.domain == domain)
        if len(res) == 1:
            return Client(**res[0])
        return None

    def update_ip(self, client: Client, ip: str, version: int):
        client_query = Query()
        ip_key = "last_ip4" if version == 4 else "last_ip6"

        self.db.update({ip_key: ip}, (client_query.uuid == client.uuid))

    @property
    def clients(self):
        return [Client(**x) for x in self.db.all()]



import json
import logging
import socket
from typing import Dict, Optional

import bcrypt
import requests
import yaml
from jsonschema import validate

anlog = logging.getLogger("anlog")


def set_loglevel(v):
    ch = logging.StreamHandler()
    if v:
        anlog.setLevel(logging.DEBUG)
        ch.setFormatter(logging.Formatter("%(levelname)s - %(msg)s"))
    else:
        anlog.setLevel(logging.INFO)
        ch.setFormatter(logging.Formatter("%(message)s"))
    anlog.addHandler(ch)


def limit_dict(dictionary, keys: tuple, translation: Optional[tuple] = None) -> Dict:
    result = {}
    if not isinstance(dictionary, Dict):
        return result
    for i, k in enumerate(keys):
        if k in dictionary:
            if translation:
                result.update({translation[i]: dictionary[k]})
            else:
                result.update({k: dictionary[k]})
    return result


def get_my_public_ip():
    r = requests.get("https://ipinfo.io/ip")
    if r.status_code == 200:
        return r.text.strip()
    return None


def hash_password(passwd):
    salt = bcrypt.gensalt()
    return bcrypt.hashpw(passwd.encode(), salt).decode()


def is_ip_record_valid(ip: str, rtype: str):
    if (ip_version(ip) == 6 and rtype != "AAAA") or (
        ip_version(ip) == 4 and rtype != "A"
    ):
        anlog.error("IP type mismatch")
        return False
    return True


def ip_version(ip: str) -> int:
    try:
        socket.inet_aton(ip)
        return 4
    except Exception:
        socket.inet_pton(socket.AF_INET6, ip)
        return 6


def record_type(ip: str):
    return "AAAA" if ip_version(ip) == 6 else "A"


config_schema = """
{
    "type": "object",
    "additionalProperties": false,
    "properties": {
        "domains": {
            "type": "array",
            "items": {
                "type": "object",
                "additionalProperties": true,
                "properties": {
                    "zone": {
                        "type": "string"
                    },
                    "provider": {
                        "type": "string"
                    }
                },
                "required": [
                    "provider",
                    "zone"
                ]
            }
        },
        "backend": {
            "type": "object",
            "additionalProperties": false,
            "properties": {
                "type": {
                    "type": "string"
                },
                "vendor": {
                    "type": "string"
                },
                "path": {
                    "type": "string"
                }
            },
            "required": [
                "path",
                "type"
            ]
        }
    },
    "required": [
        "backend",
        "domains"
    ]
}
"""


def get_or_parse_yaml(data: Dict | str) -> Dict:
    if isinstance(data, str):
        with open(data, "r") as fp:
            data = yaml.safe_load(fp.read())
            validate(data, json.loads(config_schema))
    return data



import dataclasses
from dataclasses import asdict
from time import sleep
from typing import Dict, Optional

from peewee import CharField, Database, Model, OperationalError, Proxy, SqliteDatabase
from playhouse.db_url import connect
from playhouse.shortcuts import model_to_dict

from anemoi.backends import Backend
from anemoi.client import Client
from anemoi.util import anlog, limit_dict

db_proxy = Proxy()


class BaseModel(Model):
    class Meta:
        database = db_proxy


class ClientModel(BaseModel):
    domain = CharField(max_length=253)
    uuid = CharField(max_length=36)
    secret_key = CharField()
    last_ip4 = CharField(max_length=15)
    last_ip6 = CharField(max_length=45)


def entry_to_dataclass(entry: Model, dc):
    return dc(
        **limit_dict(model_to_dict(entry), [x.name for x in dataclasses.fields(dc)])
    )


class DatabaseBackend(Backend):
    db: Database = None

    def __init__(self, config: Dict):
        allowed_db_kinds = ["sqlite", "postgres", "mysql"]

        kind = config.get("vendor")
        connection = config.get("path")
        if kind not in allowed_db_kinds:
            raise Exception(f"Backend {kind} does not exist!")
        if kind == "sqlite":
            self.db = SqliteDatabase(connection)
        elif kind in allowed_db_kinds[1:]:
            self.db = connect(connection)
        db_proxy.initialize(self.db)
        while True:
            try:
                self.db.connect()
                break
            except OperationalError:
                anlog.error("Unable to connect to database, waiting to retry...")
                sleep(5)

        self.db.create_tables([ClientModel])

    def add_client(self, client: Client):
        ClientModel.create(**asdict(client))

    def delete_client(self, client: Client):
        c = ClientModel.delete().where(ClientModel.uuid == client.uuid)
        if c.execute():
            return client.uuid
        return None

    def get_client(
        self, uuid: Optional[str] = None, domain: Optional[str] = None
    ) -> Optional[Client]:
        res: Optional[ClientModel] = None
        if uuid:
            res = ClientModel.get_or_none(ClientModel.uuid == uuid)
        elif domain:
            res = ClientModel.get_or_none(ClientModel.domain == domain)
        if res:
            return entry_to_dataclass(res, Client)
        return None

    def update_ip(self, client: Client, ip: str, version: int):
        if version == 4:
            ClientModel.update({ClientModel.last_ip4: ip}).where(
                ClientModel.uuid == client.uuid
            ).execute()
        else:
            ClientModel.update({ClientModel.last_ip6: ip}).where(
                ClientModel.uuid == client.uuid
            ).execute()

    @property
    def clients(self):
        return [entry_to_dataclass(x, Client) for x in ClientModel.select()]



from flask import Flask, abort, current_app, request

from anemoi.backends import init_backend
from anemoi.operator import ClientOperator
from anemoi.providers import Providers
from anemoi.util import anlog, get_my_public_ip, get_or_parse_yaml, record_type

app = Flask(__name__)


def setup_server(config_file):
    config = get_or_parse_yaml(config_file)
    app.config["anemoi.config"] = config

    # setup backend
    backend = init_backend(config)
    app.config["anemoi.backend"] = backend

    # setup providers
    providers = Providers(config_file=config)
    app.config["anemoi.providers"] = providers
    anlog.info("Starting anemoi...")
    return app


def get_ip():
    ip_addr = request.headers.get("cf-connecting-ip")
    if not ip_addr:
        ip_addr = request.access_route[-1]
    if ip_addr == "127.0.0.1":
        new_ip = get_my_public_ip()
        return new_ip if new_ip else ip_addr
    return ip_addr


@app.route("/")
def home():
    return "anemoi server"


@app.route("/check-in", methods=["POST", "GET"])
def check_in():
    if request.method == "POST":
        if not request.is_json:
            abort(400)
        data = dict(request.json)
    else:
        data = {
            "uuid": request.args.get("uuid"),
            "secret": request.args.get("secret"),
            "ip": request.args.get("ip"),
        }
    if not all(k in data for k in ("uuid", "secret")):
        abort(400)
    uuid = data.get("uuid")
    secret = data.get("secret")
    manually_set_ip = data.get("ip")
    co = ClientOperator(current_app.config.get("anemoi.backend"))
    res = co.validate_secret(uuid, secret)
    if not res:  # no auth, exit
        return "not changed", 200
    if client := co.backend.get_client(uuid=uuid):
        providers: Providers = current_app.config.get("anemoi.providers")
        provider = providers.get_provider(client.domain)
        ip = manually_set_ip or get_ip()
        rtype = record_type(ip)
        ips = provider.get_record_ips(client.domain)
        providers_ip_record = ""  # handle if IP changed on provider's side, then update it to the real one
        for i in ips:
            if rtype in i:  # noqa: SIM908
                providers_ip_record = i[rtype]
        if (
            co.did_ip_change(uuid, ip)
            or not providers_ip_record
            or (providers_ip_record and ip != providers_ip_record)
        ):
            co.update_ip(uuid, ip)
            if ip != providers_ip_record:
                if provider.update_record_ip(client.domain, ip, rtype=rtype):
                    msg = f"changed IP for {client.domain} to {ip}"
                else:
                    msg = f"error updating IP for {client.domain}"
            else:
                msg = f"updated IP for {client.domain} to {ip} in database"
            anlog.debug(msg)
            return msg, 200
    return "not changed", 200



from secrets import choice, token_urlsafe
from typing import Optional
from uuid import uuid4

import bcrypt

from anemoi.backends import Backend
from anemoi.client import Client
from anemoi.util import hash_password, ip_version


class ClientOperator:
    backend: Backend

    def __init__(self, backend):
        self.backend = backend

    def new_client(self, domain, firstIP4="", firstIP6="") -> tuple[Client, str]:
        aID = str(uuid4())
        # bcrypt max length is 72 characters
        # generate a random string that is URL safe/printable but also has a length between 64 and 72 bytes
        aSecret = token_urlsafe(64)[: choice(range(64, 72))]
        # firstIP can be blank
        client = Client(domain, aID, hash_password(aSecret), firstIP4, firstIP6)
        self.backend.add_client(client)
        return client, aSecret

    def validate_secret(self, uuid, secret):
        if client := self.backend.get_client(uuid=uuid):
            return bcrypt.checkpw(secret.encode(), client.secret_key.encode())
        return False

    def delete_client(
        self, uuid: Optional[str] = None, domain: Optional[str] = None
    ) -> Optional[str]:
        if client := self.backend.get_client(uuid=uuid, domain=domain):
            deleted_uuid = self.backend.delete_client(client)
            return deleted_uuid
        return None

    @property
    def clients(self):
        return self.backend.clients

    def update_ip(self, uuid: str, ip: str):
        if client := self.backend.get_client(uuid=uuid):
            self.backend.update_ip(client, ip, ip_version(ip))

    def did_ip_change(self, uuid, ip) -> bool:
        if client := self.backend.get_client(uuid=uuid):
            if ip_version(ip) == 4:
                return client.last_ip4 != ip
            else:
                return client.last_ip6 != ip
        return False



import importlib
from typing import Dict, List, Optional

from anemoi.util import anlog, get_or_parse_yaml


class Provider:
    def __init__(self, config):
        pass

    # returns list of {'A': '1.1.1.1'} objects
    def get_record_ips(self, subdomain) -> List[Dict[str, str]]:
        return []

    # returns bool of if the update succeeded or not
    def update_record_ip(self, subdomain, ip, rtype="A", **kwargs) -> bool:
        return False


class Providers:
    providers: Dict[str, Provider]

    def __init__(self, config_file):
        self.providers = {}
        conf = get_or_parse_yaml(config_file)
        for domain_config in conf.get("domains", []):
            provider_name = domain_config.get("provider", "")
            zone = domain_config.get("zone")
            provider_obj: Optional[Provider] = getattr(
                importlib.import_module(f"anemoi.providers.{provider_name}"),
                f"{provider_name.capitalize()}Provider",
            )
            if not provider_obj:
                raise ModuleNotFoundError(
                    f"No DNS provider named {provider_name.capitalize()}"
                )
            provider = provider_obj(domain_config)
            if not provider:
                anlog.error(
                    f"Unable to authenticate on {provider_name.capitalize()} for {zone}"
                )
                continue
            self.providers.update({zone: provider})

    def get_provider(self, zone) -> Optional[Provider]:
        if zone.count(".") > 1:
            zone = ".".join(zone.split(".")[-2:])
        return self.providers.get(zone)



import click
from jsonschema import ValidationError

from anemoi.backends import init_backend
from anemoi.operator import ClientOperator
from anemoi.server import setup_server
from anemoi.util import get_or_parse_yaml, set_loglevel


# walk up the ctx parent chain until you run into the desired parameter
def find_ctx_param(ctx, param: str):
    ctx_copy = ctx
    while True:
        if val := ctx_copy.params.get(param):
            return val
        ctx_copy = ctx_copy.parent


def get_config(ctx):
    config_file = find_ctx_param(ctx, "config")
    try:
        config = get_or_parse_yaml(config_file)
    except ValidationError as e:
        ctx.fail(f"Config file validation failed on {config_file}:\n{e}")
    return config


@click.group()
@click.option("-v", is_flag=True)
@click.option(
    "-c",
    "--config",
    help="Anemoi configuration file",
    default="~/.anemoi/config.yml",
    type=click.Path(exists=True),
)
@click.version_option(package_name="anemoi-dns")
def cli(v, config):
    set_loglevel(v)


@cli.command()
@click.option("-s", "--serve-host", help="IP/URL to serve anemoi", default="127.0.0.1")
@click.option("-p", "--port", help="port for server", default=9999)
@click.pass_context
def server(ctx, serve_host, port):
    app = setup_server(get_config(ctx))
    app.run(host=serve_host, port=port)


@cli.group()
def client():
    pass


@client.command(help="add new client")
@click.option("-d", "--domain", help="domain to update for this client", prompt=True)
@click.option("-i", "--ip", help="optional known starting IP")
@click.pass_context
def add(ctx, domain, ip):
    config = get_config(ctx)
    backend = init_backend(config)
    co = ClientOperator(backend)
    if not ip:
        ip = ""
    zones = [x.get("zone") for x in config.get("domains")]
    for zone in zones:
        if domain.endswith(zone):
            break
    else:
        ctx.fail(f"No zone defined for {domain}")
    client, secret = co.new_client(domain, ip)
    click.echo("\n- Client info -")
    click.echo(f"uuid: {client.uuid}")
    click.echo(f"secret: {secret}")


@client.command(help="list all clients")
@click.pass_context
def list(ctx):
    backend = init_backend(get_config(ctx))
    co = ClientOperator(backend)
    if len(co.clients) == 0:
        click.echo("No clients found")
        exit(0)
    click.echo("Clients:")
    for client in co.clients:
        click.echo(f" - {client.uuid} ({client.domain})")


@client.command(help="delete clients")
@click.option("-d", "--domain", help="domain of client to delete")
@click.option("-u", "--uuid", help="UUID of client to delete")
@click.pass_context
def delete(ctx, domain, uuid):
    if (not domain and not uuid) or (uuid and domain):
        ctx.fail("Specify one of 'uuid' or 'domain'.")
    backend = init_backend(get_config(ctx))
    co = ClientOperator(backend)
    if domain:
        res = co.delete_client(domain=domain)
    elif uuid:
        res = co.delete_client(uuid=uuid)
    if res:
        click.echo(f"Deleted client: {res}")
    else:
        click.echo("No matching clients found")


if __name__ == "__main__":
    cli()
